package dsuyasov.recognition.rulesModel.generators;

import dsuyasov.recognition.rulesModel.*;
import dsuyasov.recognition.rulesModel.sequence.MatrixChangeListener;
import dsuyasov.recognition.rulesModel.checkers.LabelChecker;
import dsuyasov.recognition.util.XMLUtil;
import dsuyasov.recognition.util.localization.Localization;
import dsuyasov.recognition.files.XMLTransferable;
import dsuyasov.recognition.files.LoadException;

import javax.swing.*;
import java.awt.FlowLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.List;
import java.util.LinkedList;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * @author Dmitriy Suyasov
 *         Date 22.05.2007
 */
public class LabelRemover implements XMLTransferable {

    private static final String SELECTED_SYMBOL_CHECKER_NODE_NAME = "selectedSymbolChecker";
    private static final String INDEX_ATTR_NAME = "index";

    private List<LabelChecker> labelCheckers;
    private List<LabelChecker> selectedLabelCheckers;

    private MatrixElement[][] matrix;
    private List<Object> symbolsToRemove;
    private MatrixChangeListener changeListener;

    public LabelRemover() {
        labelCheckers = new LinkedList<LabelChecker>();
        selectedLabelCheckers = new LinkedList<LabelChecker>();
    }

    public LabelRemover(Node node, List<LabelChecker> checkers) throws LoadException {
        labelCheckers = new LinkedList<LabelChecker>(checkers);
        List<Node> indexesNodes = XMLUtil.getNode(node, SELECTED_SYMBOL_CHECKER_NODE_NAME);
        this.selectedLabelCheckers = new LinkedList<LabelChecker>();
        for (Node indexesNode : indexesNodes) {
            int indexOfChecker = XMLUtil.getIntAttribute(indexesNode, INDEX_ATTR_NAME);
            this.selectedLabelCheckers.add(checkers.get(indexOfChecker));
        }
    }

    public void draw(JComponent parent, JDialog parentDialog) {
        parent.removeAll();

        CellularAutomata cellularAutomata = (CellularAutomata) Memory.get(Memory.CURRENT_RULE);
        //int ruleItemIndex = (Integer) Memory.get(Memory.SELECTED_RULE);
        RuleItem ruleItem = cellularAutomata.getSelectedRuleItem();//ruleItemIndex == 0 ? cellularAutomata.getDefaultRuleItem() : cellularAutomata.getRuleItems().get(ruleItemIndex - 1);
        labelCheckers = ruleItem.getCell(0, 0).getLabelCheckers();

        if (labelCheckers.size() > 0) {
            JPanel chooseCheckerPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 5));
            parent.add(chooseCheckerPanel);
            chooseCheckerPanel.add(new JLabel(Localization.get(Localization.CHOOSE_CHECKER_FIELD_TITLE) + ":"));

            String[] checkerTitles = new String[labelCheckers.size()];
            for (int iLoop = 0; iLoop < labelCheckers.size(); iLoop++) {
                LabelChecker labelChecker = labelCheckers.get(iLoop);
                checkerTitles[iLoop] = Localization.get(Localization.CHECKER_TITLE) + " " + (iLoop + 1) + ": " + labelChecker.getLocalizationTitle();
            }

            JPanel symbolCheckersListPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 5));
            parent.add(symbolCheckersListPanel);
            drawCheckers(symbolCheckersListPanel, parentDialog);

            JList selectCheckersList = new JList(checkerTitles);
            chooseCheckerPanel.add(selectCheckersList);
            JButton addChecherButton = new JButton(Localization.get(Localization.ADD_CHECKER_FOR_USAGE_BUTTON_NAME));
            addChecherButton.addActionListener(new AddCheckerListener(selectCheckersList, symbolCheckersListPanel, parentDialog));
            chooseCheckerPanel.add(addChecherButton);
            JButton clearChechersButton = new JButton(Localization.get(Localization.CLEAR_CHECKERS_BUTTON_NAME));
            clearChechersButton.addActionListener(new ClearCheckersListListener(symbolCheckersListPanel, parentDialog));
            chooseCheckerPanel.add(clearChechersButton);
        }
    }

    public void setChange() {
        changeListener.change();
    }

    private void drawCheckers(JPanel parent, JDialog parentDialog) {
        parent.removeAll();
        StringBuffer buffer = new StringBuffer(Localization.get(Localization.SYMBOLS_FOR_REMOVING_FIELD_TITLE) + ": ");
        for (LabelChecker labelChecker : selectedLabelCheckers) {
            buffer.append(buffer.length() > 0 ? ", " : "").append(labelChecker.getLocalizationTitle());
        }
        parent.add(new JLabel(buffer.toString()));
        parent.revalidate();
        parent.repaint();
        parentDialog.pack();
    }

    public void toXML(Document doc, Node rootElement) {
        for (LabelChecker labelChecker : selectedLabelCheckers) {
            XMLUtil.addAttribute(doc, XMLUtil.addNode(doc, rootElement, SELECTED_SYMBOL_CHECKER_NODE_NAME), INDEX_ATTR_NAME, String.valueOf(labelCheckers.indexOf(labelChecker)));
        }
    }

    public void initialize(MatrixElement[][] source, MatrixChangeListener listener) {
        this.matrix = source;
        for (LabelChecker labelChecker : labelCheckers) {
            labelChecker.removeListenerForRemoving();
        }
        for (LabelChecker labelChecker : selectedLabelCheckers) {
            labelChecker.addListenerForRemoving(this);
        }
        symbolsToRemove = new LinkedList<Object>();
        this.changeListener = listener;
    }

    public void removeLabels(int x, int y) {
        if (symbolsToRemove.size() > 0)
            setChange();
        for (Object symbol: symbolsToRemove) {
            this.matrix[x][y].removeLabel(symbol);
        }
        symbolsToRemove.clear();
    }

    public void addLabelToRemove(Object symbol) {
        symbolsToRemove.add(symbol);
    }

    public void clearLabelToRemove() {
        symbolsToRemove.clear();
    }

    public List<LabelChecker> getSelectedSymbolCheckers() {
        return selectedLabelCheckers;
    }

    private class ClearCheckersListListener implements ActionListener {

        private JPanel parent;
        private JDialog dialog;

        public ClearCheckersListListener(JPanel parent, JDialog dialog) {
            this.parent = parent;
            this.dialog = dialog;
        }

        public void actionPerformed(ActionEvent e) {
            selectedLabelCheckers.clear();
            drawCheckers(parent, dialog);
        }
    }

    private class AddCheckerListener implements ActionListener {

        private JList checkersList;
        private JPanel parent;
        private JDialog dialog;

        public AddCheckerListener(JList checkersList, JPanel parent, JDialog dialog) {
            this.checkersList = checkersList;
            this.parent = parent;
            this.dialog = dialog;
        }

        public void actionPerformed(ActionEvent e) {
            int selectedIndex = checkersList.getSelectedIndex();
            if (selectedIndex >= 0) {
                LabelChecker labelChecker = labelCheckers.get(selectedIndex);
                boolean contain = false;
                for (LabelChecker selectedLabelChecker : selectedLabelCheckers) {
                    if (selectedLabelChecker == labelChecker) {
                        contain = true;
                        break;
                    }
                }
                if (!contain) {
                    selectedLabelCheckers.add(labelChecker);
                    drawCheckers(parent, dialog);
                }
            }
        }
    }

}
