package dsuyasov.recognition.rulesModel.sequence;

import dsuyasov.recognition.rulesModel.*;
import dsuyasov.recognition.util.XMLUtil;
import dsuyasov.recognition.files.XMLTransferable;
import dsuyasov.recognition.files.LoadException;

import java.util.List;
import java.util.LinkedList;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * @author Dmitriy Suyasov
 *         Date 17.05.2007
 */
public class AutomataSequence implements XMLTransferable {

    public static final String TYPE_ATTR_NAME = "type";
    public static final String ELEMENT_NODE_NAME = "element";
    public static final String POSITION_ATTR_NAME = "position";
    public static final String DESCRIPTION_NODE_NAME = "description";

    private List<Processable> rules;
    private String description;

    public AutomataSequence() {
        this.rules = new LinkedList<Processable>();
        this.description = "";
    }

    public AutomataSequence(Node node) throws LoadException {
        List<Node> descriptions = XMLUtil.getNode(node, DESCRIPTION_NODE_NAME);
        if (descriptions != null && descriptions.size() > 0)
            this.description = XMLUtil.getNodeValue(descriptions.get(0));
        else
            this.description = "";
        rules = new LinkedList<Processable>();
        List<Node> nodes = XMLUtil.getNode(node, ELEMENT_NODE_NAME);
        for (Node elementNode : nodes) {
            if (XMLUtil.getAttribute(elementNode, TYPE_ATTR_NAME).equals(CellularAutomata.TYPE)) {
                rules.add(new CellularAutomata(elementNode));
            } else {
                rules.add(SpecialElement.newSpecialElementInstance(elementNode));
            }
        }
        for (SequenceDisplayable rule : rules) {
            if (rule instanceof SpecialElement) {
                ((SpecialElement) rule).updateInformation(this);
            }
        }
    }

    public AutomataSequence(AutomataSequence automataSequence) {
        rules = new LinkedList<Processable>();
        for (int iLoop = 0; iLoop < automataSequence.getRules().size(); iLoop++) {
            Processable processable =  automataSequence.getRules().get(iLoop);
            if (processable instanceof SpecialElement) {
                rules.add(SpecialElement.newInstance((SpecialElement) processable, this));
            } else {
                rules.add(processable);
            }
        }
    }

    public AutomataSequence(AutomataSequence automataSequence, int start) {
        this(automataSequence);
        for (int iLoop = start; iLoop < rules.size(); iLoop++) {
            Processable processable = rules.get(iLoop);
            if (processable instanceof SpecialElement)
                processable.initialize(this);
        }
    }

    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    public List<Processable> getRules() {
        return rules;
    }

    public void addRule(Processable rule) {
        rules.add(rule);
        for (Processable processable : rules) {
            if (processable instanceof SpecialElement)
                ((SpecialElement) processable).updateInformation(this);
        }
    }

    public Processable removeRule(int index) {
        Processable result = rules.remove(index);
        for (Processable processable : rules) {
            if (processable instanceof SpecialElement)
                ((SpecialElement) processable).updateInformationRuleRemoved(this, index);
        }
        return result;
    }

    public void moveRule(int index, int where) {
        if (index + where < 0 || index + where >= rules.size())
            return;
        Processable rule = rules.get(index);
        rules.remove(index);
        rules.add(index + where, rule);
        for (Processable processable : rules) {
            if (processable instanceof SpecialElement)
                ((SpecialElement) processable).updateInformationRulesMoved(this, (where > 0) ? index : index + where);
        }
    }

    public boolean isLastIsOutputter() {
        return rules == null || rules.isEmpty() ? false : rules.get(rules.size() - 1) instanceof OutputerElement;
    }

    public boolean canBeRemoved(int index) {
        SequenceDisplayable sequenceDisplayable = rules.get(index);
        if (sequenceDisplayable instanceof SpecialElement) {
            return ((SpecialElement) sequenceDisplayable).canBeRemoved();
        } else
            return true;
    }

    public static AutomataSequence getRulesSequence() {
        AutomataSequence automataSequence = (AutomataSequence) Memory.get(Memory.CURRENT_SEQUENCE);
        if (automataSequence == null) {
            automataSequence = new AutomataSequence();
            Memory.add(Memory.CURRENT_SEQUENCE, automataSequence);
        }
        return automataSequence;
    }

    public static void setRulesSequence(AutomataSequence automataSequence) {
        if (automataSequence != null)
            Memory.add(Memory.CURRENT_SEQUENCE, automataSequence);
    }

    public void toXML(Document doc, Node rootElement) {
        XMLUtil.addCDATASection(doc, XMLUtil.addNode(doc, rootElement, DESCRIPTION_NODE_NAME), this.description);
        for (int iLoop = 0; iLoop < rules.size(); iLoop++) {
            SequenceDisplayable sequenceDisplayable = rules.get(iLoop);
            Node node = XMLUtil.addNode(doc, rootElement, ELEMENT_NODE_NAME);
            XMLUtil.addAttribute(doc, node, POSITION_ATTR_NAME, String.valueOf(iLoop));
            sequenceDisplayable.toXML(doc, node);
        }
    }

    public void process(MatrixElement[][] matrixElements) {

        // initialization
        for (Processable rule : rules) {
            rule.initialize(this);
        }

        // start
        processingAll(new ForProcess(matrixElements, 0, this));
    }

    public void initializeProcessing(MatrixElement[][] matrixElements) {
        // initialization
        for (Processable rule : rules) {
            rule.initialize(this);
        }

        List<ForProcess> elements = new LinkedList<ForProcess>();
        elements.add(new ForProcess(matrixElements, 0, this));
        Memory.add(Memory.PROCESSING_ELEMENTS, elements);
        Memory.dropIncreaser();
    }

    public void processStep() {
        List<ForProcess> forProcessing = (List<ForProcess>) Memory.get(Memory.PROCESSING_ELEMENTS);
        forProcessing.addAll(0, processNextElement(forProcessing.remove(0), true));
        Memory.add(Memory.PROCESSING_ELEMENTS, forProcessing);
    }

    public void processAutomata() {
        List<ForProcess> forProcessing = (List<ForProcess>) Memory.get(Memory.PROCESSING_ELEMENTS);
        forProcessing.addAll(0, processNextElement(forProcessing.remove(0), false));
        Memory.add(Memory.PROCESSING_ELEMENTS, forProcessing);
    }

    public void processAll() {
        for (ForProcess forProcess : (List<ForProcess>) Memory.get(Memory.PROCESSING_ELEMENTS)) {
            processingAll(forProcess);
        }
        Memory.add(Memory.PROCESSING_ELEMENTS, new LinkedList<ForProcess>());
    }

    private void processingAll(ForProcess element) {
        List<ForProcess> elements = new LinkedList<ForProcess>();
        elements.add(element); 
        while (elements.size() == 1) {
            elements = processNextElement(elements.get(0), false);
        }
        if (!elements.isEmpty()) {
            for (ForProcess forProcess : elements) {
                processingAll(forProcess);
            }
        }
    }

    private List<ForProcess> processNextElement(ForProcess forProcess, boolean step) {
        List<ForProcess> nextElements = new LinkedList<ForProcess>();
        Integer ruleIndex = forProcess.getNext();
        Processable processable = forProcess.getRuleSequence().getRules().get(ruleIndex);

        ProcessResult result;
        if (step)
            result = processable.processStep(forProcess.getElements());
        else
            result = processable.process(forProcess.getElements());

        Integer nextIndex = processable.getNext(ruleIndex);
        if (result.getElements().size() == 1) {
            nextElements.add(new ForProcess(result.getElements().get(0), nextIndex, forProcess.getRuleSequence()));
        } else {
            for (MatrixElement[][] matrixElements1 : result.getElements()) {
                nextElements.add(new ForProcess(matrixElements1, nextIndex, new AutomataSequence(forProcess.getRuleSequence(), nextIndex)));
            }
        }
        return nextElements;
    }

    public class ForProcess {
        private MatrixElement[][] elements;
        private int next;
        private AutomataSequence ruleSequence;

        public ForProcess(MatrixElement[][] elements, int next, AutomataSequence ruleSequence) {
            this.elements = elements;
            this.next = next;
            this.ruleSequence = ruleSequence;
        }

        public AutomataSequence getRuleSequence() {
            return ruleSequence;
        }

        public MatrixElement[][] getElements() {
            return elements;
        }

        public int getNext() {
            return next;
        }
    }

}
