package dsuyasov.recognition.rulesModel.sequence;

import dsuyasov.recognition.ui.helper.panels.SelectLabelsPanel;
import dsuyasov.recognition.ui.helper.LabelsChanges;
import dsuyasov.recognition.rulesModel.labels.Labels;
import dsuyasov.recognition.files.LoadException;
import dsuyasov.recognition.rulesModel.MatrixElement;
import dsuyasov.recognition.rulesModel.ProcessResult;
import dsuyasov.recognition.rulesModel.checkers.GeneratedLabelChecker;
import dsuyasov.recognition.util.XMLUtil;
import dsuyasov.recognition.util.localization.Localization;

import javax.swing.*;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * @author Dmitriy Suyasov
 *         Date 17.05.2007
 */
public class SplitIntoLabelsElement extends SpecialElement implements LabelsChanges {

    public static final String TITLE = "Split into symbols";
    public static final String LOCALIZATION_TITLE = Localization.get(Localization.SPLIT_INTO_SYMBOLS_TITLE);

    public static final String GENERATED_SYMBOLS_ATTR_NAME = "generatedSymbols";
    public static final String SYMBOL_NODE_NAME = "symbol";

    private List<Labels> labels;
    private boolean generatedLabels;

    private JDialog dialog;

    public SplitIntoLabelsElement() {
        this.labels = new LinkedList<Labels>();
        this.generatedLabels = true;
    }

    public SplitIntoLabelsElement(Node node) throws LoadException {
        this.generatedLabels = XMLUtil.getBooleanAttribute(node, GENERATED_SYMBOLS_ATTR_NAME);
        this.labels = new LinkedList<Labels>();
        List<Node> symbolNodes = XMLUtil.getNode(node, SYMBOL_NODE_NAME);
        for (Node symbolNode : symbolNodes) {
            this.labels.add(Labels.newLabelInstance(symbolNode));
        }
    }

    public SplitIntoLabelsElement(SplitIntoLabelsElement splitIntoLabelsElement, AutomataSequence automataSequence) {
        this.generatedLabels = splitIntoLabelsElement.isGeneratedLabels();
        this.labels = splitIntoLabelsElement.getLabels();
    }

    public boolean isGeneratedLabels() {
        return generatedLabels;
    }

    public List<Labels> getLabels() {
        return labels;
    }

    public String getTitle() {
        return TITLE;
    }

    public String getLocalizationTitle() {
        return LOCALIZATION_TITLE;
    }

    public void draw(Graphics g, Rectangle rect, int xTextOffset, int yTextOffset) {
        g.setColor(Color.BLACK);
        g.drawLine(rect.x, rect.y, rect.x + rect.width, rect.y);
        g.drawLine(rect.x, rect.y + rect.height, rect.x + rect.width, rect.y + rect.height);
        g.drawLine(rect.x + rect.width / 10, rect.y, rect.x, rect.y + rect.height);
        g.drawLine(rect.x, rect.y + rect.height, rect.x, rect.y + rect.height * 2 / 3);
        g.drawLine(rect.x + rect.width * 9 / 10, rect.y, rect.x + rect.width, rect.y + rect.height);
        g.drawLine(rect.x + rect.width, rect.y + rect.height, rect.x + rect.width, rect.y + rect.height * 2 / 3);
        super.draw(g, rect, rect.width / 10, yTextOffset);
    }

    public String getDescription() {
        return Localization.get(Localization.SPLITTING_DESCRIPTION);
    }

    protected void drawSettingsFromCleanList(JDialog dialog, JPanel panel, AutomataSequence automataSequence) {
        this.dialog = dialog;
        JPanel labelsSettingsPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 5));
        panel.add(labelsSettingsPanel);

        labelsSettingsPanel.add(new JLabel(Localization.get(Localization.LABELS_TO_USE_TITLE)));
        JPanel labelsPanel = new SelectLabelsPanel(labels, this);
        labelsSettingsPanel.add(labelsPanel);

        JPanel generatedLabelsSettingsPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 5, 5));
        panel.add(generatedLabelsSettingsPanel);
        JCheckBox generatedLabelsCheck = new JCheckBox(Localization.get(Localization.USE_GENERATED_LABELS_TITLE), this.generatedLabels);
        generatedLabelsCheck.addChangeListener(new GeneratedLabelsChangeListener());
        generatedLabelsSettingsPanel.add(generatedLabelsCheck);
    }

    public boolean confirmCorrectness() {
        return generatedLabels || labels.size() > 0;
    }

    public String getErrorMessage() {
        return (!generatedLabels && labels.size() == 0) ? Localization.get(Localization.CHECK_SELECTED_LABELS_MESSAGE) : "";
    }

    public boolean canBeRemoved() {
        return true;
    }

    public void updateInformation(AutomataSequence automataSequence) {

    }

    public void updateLabels(List<Labels> symbols) {
        this.labels = new LinkedList<Labels>(symbols);
        dialog.pack();
    }

    public void toXML(Document doc, Node rootElement) {
        super.toXML(doc, rootElement);
        XMLUtil.addAttribute(doc, rootElement, GENERATED_SYMBOLS_ATTR_NAME, String.valueOf(generatedLabels));
        for (Labels label : labels) {
            label.toXML(doc, XMLUtil.addNode(doc, rootElement, SYMBOL_NODE_NAME));
        }
    }

    public void initialize(AutomataSequence automataSequence) {
    }

    public ProcessResult process(MatrixElement[][] source) {
        Map<Integer, Integer> minX = new HashMap<Integer, Integer>();
        Map<Integer, Integer> maxX = new HashMap<Integer, Integer>();
        Map<Integer, Integer> minY = new HashMap<Integer, Integer>();
        Map<Integer, Integer> maxY = new HashMap<Integer, Integer>();

        for (int y = 0; y < source[0].length; y++) {
            for (int x = 0; x < source.length; x++) {
                for (Object label : source[x][y].getLabels()) {
                    if (generatedLabels && (label instanceof Integer)) {
                        addIntoMap(minX, maxX, minY, maxY, x, y, (Integer) label);
                    }
                }
            }
        }

        ProcessResult processResult = new ProcessResult();
        List<Integer> aList = new LinkedList<Integer>(minX.keySet());
        Collections.sort(aList, new LabelsComparator(minX, maxX, minY, maxY));

        for (Integer key : aList) {
            Integer xStart = minX.get(key);
            Integer xEnd = maxX.get(key);
            Integer yStart = minY.get(key);
            Integer yEnd = maxY.get(key);
            MatrixElement[][] me = new MatrixElement[xEnd - xStart + 1][yEnd - yStart + 1];
            for (int x = xStart; x <= xEnd; x++) {
                for (int y = yStart; y <= yEnd; y++) {
                    me[x - xStart][y - yStart] = GeneratedLabelChecker.checkTheGeneratedSymbol(source[x][y].getLabels(), key) ? source[x][y] : new MatrixElement(Color.WHITE);
                }
            }
            processResult.add(me);
        }

        return processResult;
    }

    private class LabelsComparator implements Comparator<Integer> {
        private Map<Integer, Integer> minX;
        private Map<Integer, Integer> maxX;
        private Map<Integer, Integer> minY;
        private Map<Integer, Integer> maxY;

        public LabelsComparator(Map<Integer, Integer> minX, Map<Integer, Integer> maxX, Map<Integer, Integer> minY, Map<Integer, Integer> maxY) {
            this.minX = minX;
            this.maxX = maxX;
            this.minY = minY;
            this.maxY = maxY;
        }

        public int compare(Integer o1, Integer o2) {
            if (minY.get(o2) > maxY.get(o1))
                return -1;
            else if (minY.get(o1) > maxY.get(o2))
                return 1;
            else
                return minX.get(o1).compareTo(minX.get(o2));
        }
    }

    private void addIntoMap(Map<Integer, Integer> minX, Map<Integer, Integer> maxX, Map<Integer, Integer> minY,
                            Map<Integer, Integer> maxY, int x, int y, Integer key) {
        minX.put(key, min(x, minX.get(key)));
        maxX.put(key, max(x, maxX.get(key)));
        minY.put(key, min(y, minY.get(key)));
        maxY.put(key, max(y, maxY.get(key)));
    }

    private int min(Integer a, Integer b) {
        return b == null ? a : Math.min(a, b);
    }
    private int max(Integer a, Integer b) {
        return b == null ? a : Math.max(a, b);
    }

    private class GeneratedLabelsChangeListener implements ChangeListener {
        public void stateChanged(ChangeEvent e) {
            generatedLabels = ((JCheckBox) e.getSource()).isSelected();
        }
    }
}
